@@ -1,7 +1,8 @@
+# flake8: noqa: F401
 from rpyc.core.stream import SocketStream, TunneledSocketStream, PipeStream
 from rpyc.core.channel import Channel
 from rpyc.core.protocol import Connection
 from rpyc.core.netref import BaseNetref
-from rpyc.core.async import AsyncResult, AsyncResultTimeout
-from rpyc.core.service import Service, VoidService, SlaveService
+from rpyc.core.async_ import AsyncResult, AsyncResultTimeout
+from rpyc.core.service import Service, VoidService, SlaveService, MasterService, ClassicService
 from rpyc.core.vinegar import GenericException
