@@ -5,38 +5,47 @@ and various platforms (posix/windows)
 import sys
 import time
 
-is_py3k = (sys.version_info[0] >= 3)
+is_py_3k = (sys.version_info[0] >= 3)
+is_py_gte38 = is_py_3k and (sys.version_info[1] >= 8)
+is_py_gte37 = is_py_3k and (sys.version_info[1] >= 7)
 
-if is_py3k:
+
+if is_py_3k:
     exec("execute = exec")
+
     def BYTES_LITERAL(text):
         return bytes(text, "utf8")
     maxint = sys.maxsize
 else:
     exec("""def execute(code, globals = None, locals = None):
                 exec code in globals, locals""")
+
     def BYTES_LITERAL(text):
         return text
     maxint = sys.maxint
 
 try:
-    from struct import Struct #@UnusedImport
+    from struct import Struct  # noqa: F401
 except ImportError:
     import struct
+
     class Struct(object):
         __slots__ = ["format", "size"]
+
         def __init__(self, format):
             self.format = format
             self.size = struct.calcsize(format)
+
         def pack(self, *args):
             return struct.pack(self.format, *args)
+
         def unpack(self, data):
             return struct.unpack(self.format, data)
 
 try:
     from cStringIO import StringIO as BytesIO
 except ImportError:
-    from io import BytesIO #@UnusedImport
+    from io import BytesIO  # noqa: F401
 
 try:
     next = next
@@ -47,7 +56,7 @@ except NameError:
 try:
     import cPickle as pickle
 except ImportError:
-    import pickle #@UnusedImport
+    import pickle  # noqa: F401
 
 try:
     callable = callable
@@ -59,6 +68,7 @@ try:
     import select as select_module
 except ImportError:
     select_module = None
+
     def select(*args):
         raise ImportError("select not supported on this platform")
 else:
@@ -68,12 +78,14 @@ else:
     else:
         from select import select
 
+
 def get_exc_errno(exc):
     if hasattr(exc, "errno"):
         return exc.errno
     else:
         return exc[0]
 
+
 if select_module:
     select_error = select_module.error
 else:
@@ -83,6 +95,7 @@ if hasattr(select_module, "poll"):
     class PollingPoll(object):
         def __init__(self):
             self._poll = select_module.poll()
+
         def register(self, fd, mode):
             flags = 0
             if "r" in mode:
@@ -98,12 +111,14 @@ if hasattr(select_module, "poll"):
                 flags |= select_module.POLLHUP | select_module.POLLNVAL | POLLRDHUP
             self._poll.register(fd, flags)
         modify = register
+
         def unregister(self, fd):
             self._poll.unregister(fd)
-        def poll(self, timeout = None):
+
+        def poll(self, timeout=None):
             if timeout:
                 # the real poll takes milliseconds while we have seconds here
-                timeout = 1000*timeout
+                timeout = 1000 * timeout
             events = self._poll.poll(timeout)
             processed = []
             for fd, evt in events:
@@ -127,16 +142,19 @@ else:
         def __init__(self):
             self.rlist = set()
             self.wlist = set()
+
         def register(self, fd, mode):
             if "r" in mode:
                 self.rlist.add(fd)
             if "w" in mode:
                 self.wlist.add(fd)
         modify = register
+
         def unregister(self, fd):
             self.rlist.discard(fd)
             self.wlist.discard(fd)
-        def poll(self, timeout = None):
+
+        def poll(self, timeout=None):
             if not self.rlist and not self.wlist:
                 time.sleep(timeout)
                 return []  # need to return an empty array in this case
@@ -157,10 +175,11 @@ def with_metaclass(meta, *bases):
             return meta(name, bases, d)
     return type.__new__(metaclass, 'temporary_class', (), {})
 
+
 if sys.version_info >= (3, 3):
-    TimeoutError = TimeoutError
+    TimeoutError = TimeoutError  # noqa: F821
 else:
-    class TimeoutError(Exception):
+    class TimeoutError(Exception):  # noqa: F821
         pass
 
 if sys.version_info >= (3, 2):
