@@ -5,7 +5,7 @@ Requires [plumbum](http://plumbum.readthedocs.org/)
 """
 from __future__ import with_statement
 import sys
-import socket
+import socket  # noqa: F401
 from rpyc.lib.compat import BYTES_LITERAL
 from rpyc.core.service import VoidService
 from rpyc.core.stream import SocketStream
@@ -13,6 +13,7 @@ import rpyc.utils.factory
 import rpyc.utils.classic
 try:
     from plumbum import local, ProcessExecutionError, CommandNotFound
+    from plumbum.commands.base import BoundCommand
     from plumbum.path import copy
 except ImportError:
     import inspect
@@ -64,6 +65,7 @@ finally:
     thd.join(2)
 """
 
+
 class DeployedServer(object):
     """
     Sets up a temporary, short-lived RPyC deployment on the given remote machine. It will:
@@ -85,7 +87,8 @@ class DeployedServer(object):
     :param extra_setup: any extra code to add to the script
     """
 
-    def __init__(self, remote_machine, server_class = "rpyc.utils.server.ThreadedServer", extra_setup = "", python_executable=None):
+    def __init__(self, remote_machine, server_class="rpyc.utils.server.ThreadedServer",
+                 extra_setup="", python_executable=None):
         self.proc = None
         self.tun = None
         self.remote_machine = remote_machine
@@ -98,8 +101,11 @@ class DeployedServer(object):
 
         script = (tmp / "deployed-rpyc.py")
         modname, clsname = server_class.rsplit(".", 1)
-        script.write(SERVER_SCRIPT.replace("$MODULE$", modname).replace("$SERVER$", clsname).replace("$EXTRA_SETUP$", extra_setup))
-        if python_executable:
+        script.write(SERVER_SCRIPT.replace("$MODULE$", modname).replace(
+            "$SERVER$", clsname).replace("$EXTRA_SETUP$", extra_setup))
+        if isinstance(python_executable, BoundCommand):
+            cmd = python_executable
+        elif python_executable:
             cmd = remote_machine[python_executable]
         else:
             major = sys.version_info[0]
@@ -115,7 +121,7 @@ class DeployedServer(object):
             if not cmd:
                 cmd = remote_machine.python
 
-        self.proc = cmd.popen(script, new_session = True)
+        self.proc = cmd.popen(script, new_session=True)
 
         line = ""
         try:
@@ -138,8 +144,10 @@ class DeployedServer(object):
 
     def __del__(self):
         self.close()
+
     def __enter__(self):
         return self
+
     def __exit__(self, t, v, tb):
         self.close()
 
@@ -147,15 +155,26 @@ class DeployedServer(object):
         if self.proc is not None:
             try:
                 self.proc.terminate()
+                self.proc.communicate()
             except Exception:
                 pass
             self.proc = None
         if self.tun is not None:
             try:
+                self.tun._session.proc.terminate()
+                self.tun._session.proc.communicate()
                 self.tun.close()
             except Exception:
                 pass
             self.tun = None
+        if self.remote_machine is not None:
+            try:
+                self.remote_machine._session.proc.terminate()
+                self.remote_machine._session.proc.communicate()
+                self.remote_machine.close()
+            except Exception:
+                pass
+            self.remote_machine = None
         if self._tmpdir_ctx is not None:
             try:
                 self._tmpdir_ctx.__exit__(None, None, None)
@@ -170,7 +189,7 @@ class DeployedServer(object):
         else:
             return SocketStream._connect("localhost", self.local_port)
 
-    def connect(self, service = VoidService, config = {}):
+    def connect(self, service=VoidService, config={}):
         """Same as :func:`~rpyc.utils.factory.connect`, but with the ``host`` and ``port``
         parameters fixed"""
         return rpyc.utils.factory.connect_stream(
@@ -188,21 +207,27 @@ class MultiServerDeployment(object):
     An 'aggregate' server deployment to multiple SSH machine. It deploys RPyC to each machine
     separately, but lets you manage them as a single deployment.
     """
-    def __init__(self, remote_machines, server_class = "rpyc.utils.server.ThreadedServer"):
+
+    def __init__(self, remote_machines, server_class="rpyc.utils.server.ThreadedServer"):
         self.remote_machines = remote_machines
         # build the list incrementally, so we can clean it up if we have an exception
         self.servers = [DeployedServer(mach, server_class) for mach in remote_machines]
 
     def __del__(self):
         self.close()
+
     def __enter__(self):
         return self
+
     def __exit__(self, t, v, tb):
         self.close()
+
     def __iter__(self):
         return iter(self.servers)
+
     def __len__(self):
         return len(self.servers)
+
     def __getitem__(self, index):
         return self.servers[index]
 
@@ -211,9 +236,10 @@ class MultiServerDeployment(object):
             s = self.servers.pop(0)
             s.close()
 
-    def connect_all(self, service = VoidService, config = {}):
+    def connect_all(self, service=VoidService, config={}):
         """connects to all deployed servers; returns a list of connections (order guaranteed)"""
         return [s.connect(service, config) for s in self.servers]
+
     def classic_connect_all(self):
         """connects to all deployed servers using classic_connect; returns a list of connections (order guaranteed)"""
         return [s.classic_connect() for s in self.servers]
