@@ -37,20 +37,10 @@ class ClassicMode(unittest.TestCase):
         self.assertEqual(list(bi), list(range(10000)))
 
     def test_classic(self):
-        print( self.conn.modules.sys )
-        print( self.conn.modules["xml.dom.minidom"].parseString("<a/>") )
         self.conn.execute("x = 5")
         self.assertEqual(self.conn.namespace["x"], 5)
         self.assertEqual(self.conn.eval("1+x"), 6)
 
-    def test_isinstance(self):
-        x = self.conn.builtin.list((1,2,3,4))
-        print( x )
-        print( type(x) )
-        print( x.__class__ )
-        self.assertTrue(isinstance(x, list))
-        self.assertTrue(isinstance(x, rpyc.BaseNetref))
-
     def test_mock_connection(self):
         from rpyc.utils.classic import MockClassicConnection
         import sys
