@@ -1,4 +1,3 @@
-import time
 import rpyc
 from rpyc.utils.server import OneShotServer
 import unittest
@@ -15,20 +14,21 @@ class Test_OneShotServer(unittest.TestCase):
     def setUp(self):
         self.server = OneShotServer(MyService, port=18878, auto_register=False)
         self.server.logger.quiet = False
-        self.server._start_in_thread()
+        self.thd = self.server._start_in_thread()
 
     def tearDown(self):
         self.server.close()
+        self.thd.join()
 
     def test_server_stops(self):
         conn = rpyc.connect("localhost", port=18878)
         self.assertEqual("bar", conn.root.foo())
         conn.close()
+        while not self.server._closed:
+            pass
+        self.assertTrue(self.server._closed)
         with self.assertRaises(Exception):
-            for i in range(3):
-                conn = rpyc.connect("localhost", port=18878)
-                conn.close()
-                time.sleep()
+            conn = rpyc.connect("localhost", port=18878)
 
 
 if __name__ == "__main__":
