@@ -2,7 +2,6 @@ import os
 import tempfile
 import shutil
 import unittest
-from nose import SkipTest
 import rpyc
 
 
@@ -32,29 +31,30 @@ class Test_Remoting(unittest.TestCase):
 
         shutil.rmtree(base)
 
+    @unittest.skip("TODO: upload a package and a module")
     def test_distribution(self):
-        raise SkipTest("TODO: upload a package and a module")
+        pass
 
+    @unittest.skip("Requires manual testing atm")
     def test_interactive(self):
-        raise SkipTest("Need to be manually")
-        print( "type Ctrl+D to exit (Ctrl+Z on Windows)" )
+        print("type Ctrl+D to exit (Ctrl+Z on Windows)")
         rpyc.classic.interact(self.conn)
 
+    @unittest.skip("Requires manual testing atm")
     def test_post_mortem(self):
-        raise SkipTest("Need to be manually")
         try:
             self.conn.modules.sys.path[100000]
         except IndexError:
-            print( "type 'q' to exit" )
+            print("type 'q' to exit")
             rpyc.classic.pm(self.conn)
             raise
         else:
             self.fail("expected an exception")
 
     def test_migration(self):
-        l = rpyc.classic.obtain(self.conn.modules.sys.path)
-        self.assertTrue(type(l) is list)
-        rl = rpyc.classic.deliver(self.conn, l)
+        path_list = rpyc.classic.obtain(self.conn.modules.sys.path)
+        self.assertTrue(type(path_list) is list)
+        rl = rpyc.classic.deliver(self.conn, path_list)
         self.assertTrue(isinstance(rl, rpyc.BaseNetref))
 
 
