@@ -1,16 +1,12 @@
+from rpyc.lib.compat import BYTES_LITERAL
+from rpyc.core.stream import PipeStream, NamedPipeStream
+import rpyc
 import sys
 import time
 import unittest
 
-from nose import SkipTest
-if sys.platform != "win32":
-    raise SkipTest("Requires windows")
-
-import rpyc
-from rpyc.core.stream import PipeStream, NamedPipeStream
-from rpyc.lib.compat import BYTES_LITERAL
-
 
+@unittest.skipIf(sys.platform != "win32", "Requires windows")
 class Test_Pipes(unittest.TestCase):
     def test_basic_io(self):
         p1, p2 = PipeStream.create_pair()
@@ -39,10 +35,11 @@ class Test_Pipes(unittest.TestCase):
         server_thread.join()
 
 
+@unittest.skipIf(sys.platform != "win32", "Requires windows")
 class Test_NamedPipe(object):
     def setUp(self):
         self.pipe_server_thread = rpyc.spawn(self.pipe_server)
-        time.sleep(1) # make sure server is accepting already
+        time.sleep(1)  # make sure server is accepting already
         self.np_client = NamedPipeStream.create_client("floop")
         self.client = rpyc.connect_stream(self.np_client)
 
